import cattrs.strategies

from ._arbitrary_roi import ArbitraryROI
from ._rectangular_roi import RectangularROI
from ._roi import ROI
from ._rotated_rectangular_roi import RotatedRectangularROI

converter = cattrs.Converter(
    unstruct_collection_overrides={tuple: tuple},
)
"""A converter that can (un)structure ROI objects.

.. Warning::
    Only subclasses of ROI that are defined in this module have their types correctly
    unstructured.
"""

cattrs.strategies.include_subclasses(
    ROI, converter, (RectangularROI, ArbitraryROI, RotatedRectangularROI)
)



"""Define devices that outputs a sequence of values."""

from . import channel_commands
from . import compilation
from . import instructions
from . import trigger
from ._controller import SequencerController
from ._converter import converter
from ._proxy import SequencerProxy
from ._time_step import TimeStep
from .compilation import SequencerCompiler
from .configuration import (
    SequencerConfiguration,
    ChannelConfiguration,
    DigitalChannelConfiguration,
    AnalogChannelConfiguration,
)
from .runtime import Sequencer

__all__ = [
    "Sequencer",
    "SequencerConfiguration",
    "ChannelConfiguration",
    "DigitalChannelConfiguration",
    "AnalogChannelConfiguration",
    "SequencerProxy",
    "SequencerController",
    "channel_commands",
    "instructions",
    "converter",
    "TimeStep",
    "trigger",
    "compilation",
]



import contextlib
from typing import TypeVar

from PySide6.QtWidgets import QWidget

T = TypeVar("T", bound=QWidget)


@contextlib.contextmanager
def temporary_widget(widget: T) -> T:
    """Context manager that deletes a widget when the context is exited.

    When a widget is created for a lifetime shorter than the lifetime of its parent,
    it is necessary to delete it manually to avoid memory leaks.
    This is easily forgotten, so this context manager can be used to ensure that
    the widget is deleted when it is no longer needed.

    Example:

        .. code-block:: python

                with temporary_widget(QWidget(parent=parent)) as widget:
                    widget.show()
                    widget.do_something()
    """

    try:
        yield widget
    finally:
        widget.deleteLater()



import warnings

from caqtus.types.image import (
    Width,
    Height,
)
from caqtus.types.image.roi import (
    ArbitraryROI,
    RectangularROI,
    RotatedRectangularROI,
    ROI,
)

warnings.warn(
    "caqtus.utils.roi is deprecated, use caqtus.types.image instead.",
    DeprecationWarning,
)


__all__ = [
    "ArbitraryROI",
    "RectangularROI",
    "RotatedRectangularROI",
    "ROI",
    "Width",
    "Height",
]



from collections.abc import Iterable, Callable, Iterator
from typing import TypeVar, SupportsInt

T = TypeVar("T")


def iterate_in_order(
    iterable: Iterable[T], index: Callable[[T], SupportsInt]
) -> Iterator[T]:
    """Iterate over an iterable in order of the index of its elements.

    The indices returned by the index function must be unique and consecutive, starting from 0.

    This function can be useful when you need to process elements in a specific order even if they are generated in a
    random order.

    Raises:
        ValueError: if the index of an element is not unique or if there is a gap in the indices.
    """

    result = {}
    current_index = 0
    for value in iterable:
        value_index = int(index(value))
        if value_index in result:
            raise ValueError(f"Index {value_index} appears twice")
        else:
            result[value_index] = value
        if current_index in result:
            yield result.pop(current_index)
            current_index += 1
    while current_index in result:
        yield result.pop(current_index)
        current_index += 1
    if result:
        raise ValueError(f"No value corresponds to index {current_index}")



import contextlib
from collections.abc import Generator

from PySide6.QtCore import QObject


@contextlib.contextmanager
def block_signals(obj: QObject) -> Generator[None, None, None]:
    """Context manager to block signals from a QObject."""

    obj.blockSignals(True)
    yield
    obj.blockSignals(False)



import abc
from collections.abc import Mapping, Set

from PySide6.QtWidgets import QStyledItemDelegate

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.gui.qtutil import QABCMeta


class TimeLaneDelegate(QStyledItemDelegate, metaclass=QABCMeta):
    """A delegate to display and edit the cells of a time lane."""

    def set_device_configurations(
        self, device_configurations: Mapping[DeviceName, DeviceConfiguration]
    ) -> None:
        """Set the device configurations.

        This method is called when the device configurations are updated.
        The delegate should use this information to update the appearance of the lane
        cells or editor if they depend on the device configurations.
        """

        pass

    def set_parameter_names(self, parameter_names: Set[str]) -> None:
        """Update the name of the parameters used in the sequence.

        This method is called when the parameter names are updated.
        The delegate should use this information to update the appearance of the lane
        cells or editors if they depend on the parameter names.
        """

        pass



from typing import TypeVar, Callable

from attr import AttrsInstance
from cattrs.gen import (
    AttributeOverride,
    make_dict_unstructure_fn,
    make_dict_structure_fn,
)

from .converters import converters

_T = TypeVar("_T", bound=AttrsInstance)


def customize(**kwargs: AttributeOverride) -> Callable[[type[_T]], type[_T]]:
    def decorator(cls: type[_T]) -> type[_T]:
        for converter in converters.values():
            unstructure_hook = make_dict_unstructure_fn(cls, converter, **kwargs)
            converter.register_unstructure_hook(cls, unstructure_hook)

            structure_hook = make_dict_structure_fn(cls, converter, **kwargs)
            converter.register_structure_hook(cls, structure_hook)

        return cls

    return decorator



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Protocol, TYPE_CHECKING

from returns.result import Result

from ._path import PureSequencePath

if TYPE_CHECKING:
    from ._sequence_collection import PathIsSequenceError


class PathHierarchy(Protocol):
    """A file-system-like object that exists inside a session.

    This object provides methods that can be used to create, delete and check for the
    existence of sequences.
    """

    @abstractmethod
    def does_path_exists(self, path: PureSequencePath) -> bool:
        """Check if the path exists in the session.

        Args:
            path: the path to check for existence.

        Returns:
            True if the path exists in the session. False otherwise.
        """

        raise NotImplementedError

    @abstractmethod
    def create_path(
        self, path: PureSequencePath
    ) -> Result[list[PureSequencePath], PathIsSequenceError]:
        """Create the path in the session and its parent paths if they do not exist.

        Args:
            path: the path to create.

        Returns:
            A list of the paths that were created if the path was created successfully
            or PathIsSequenceError if the path or one of its ancestors is a sequence.
            No path is created if any of the ancestors is a sequence.
            The list is ordered from parent to child.
        """

        raise NotImplementedError

    @abstractmethod
    def delete_path(self, path: PureSequencePath, delete_sequences: bool = False):
        """Delete the path and all its descendants.

        Warnings:
            If delete_sequences is True, all sequences in the path will be deleted.

        Args:
            path: The path to delete. Descendants will be deleted recursively.
            delete_sequences: If False, raise an error if the path or one of its
            children is a sequence.

        Raises:
            PathNotFoundError: If the path does not exist.
            PathIsSequenceError: If the path or one of its children is a sequence and
            delete_sequence is False.
        """

        raise NotImplementedError

    @abstractmethod
    def get_children(
        self, path: PureSequencePath
    ) -> Result[set[PureSequencePath], PathNotFoundError | PathIsSequenceError]:
        """Get the children of the path."""

        raise NotImplementedError

    @abstractmethod
    def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Result[datetime, PathNotFoundError | PathIsRootError]:
        """Get the creation date of the path.

        Args:
            path: the path to get the creation date for.

        Returns:
            The creation date of the path.
        """

        raise NotImplementedError

    @abstractmethod
    def get_all_paths(self) -> set[PureSequencePath]:
        """Get all the paths in the session.

        Returns:
            A set of all the paths in the session.
        """

        raise NotImplementedError

    @abstractmethod
    def update_creation_date(self, path: PureSequencePath, date: datetime) -> None:
        """Update the creation date of the path.

        This method is meant to be used for maintenance purposes only, such as when
        copying sequences from one session to another.

        Args:
            path: the path to update the creation date for.
            date: the new creation date.
        """

        raise NotImplementedError


class PathError(RuntimeError):
    """Base class for all path errors."""

    pass


class PathNotFoundError(PathError):
    """Raised when a path is not found in the session."""

    pass


class PathIsRootError(PathError):
    """Raised when an invalid operation is performed on the root path."""

    pass


class PathHasChildrenError(PathError):
    """Raised when an invalid operation is performed on a path that has children."""

    pass



from sqlite3 import Connection as SQLite3Connection
from typing import Self

import attrs
import sqlalchemy
import sqlalchemy.orm
import yaml
from sqlalchemy import event, Engine, create_engine, URL
from sqlalchemy.ext.asyncio import create_async_engine

from ._async_session import AsyncSQLExperimentSession, ThreadedAsyncSQLExperimentSession
from ._experiment_session import SQLExperimentSession
from ._serializer import SerializerProtocol
from ._table_base import create_tables
from .._session_maker import ExperimentSessionMaker


# We need to enable foreign key constraints for sqlite databases and not for other
# types of databases.
@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection, connection_record):
    if isinstance(dbapi_connection, SQLite3Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()


class SQLExperimentSessionMaker(ExperimentSessionMaker):
    """Used to access experiment storage were the data are stored in a SQL database.

    This session maker can create session that connects to a database using sqlalchemy.

    This object is pickleable and can be passed to other processes, assuming that the
    database referenced by the engine is accessible from the other processes.
    In particular, in-memory sqlite databases are not accessible from other processes.

    Args:
        engine: This is used by the sessions to connect to the database.
            See sqlalchemy documentation for more information on how to create an
            engine.
        serializer: This is used to convert user defined objects to a JSON format that
            can be stored in the database.

    """

    def __init__(
        self,
        serializer: SerializerProtocol,
        engine: sqlalchemy.Engine,
        async_engine: sqlalchemy.ext.asyncio.AsyncEngine,
    ) -> None:
        self._engine = engine
        self._async_engine = async_engine
        self._session_maker = sqlalchemy.orm.sessionmaker(self._engine)
        self._async_session_maker = sqlalchemy.ext.asyncio.async_sessionmaker(
            self._async_engine
        )
        self._serializer = serializer

    def create_tables(self) -> None:
        """Create the tables in the database.

        This method is useful the first time the database is created.
        It will create missing tables and ignore existing ones.
        """

        create_tables(self._engine)

    def __call__(self) -> SQLExperimentSession:
        """Create a new ExperimentSession with the engine used at initialization."""

        return SQLExperimentSession(
            self._session_maker.begin(),
            self._serializer,
        )

    def async_session(self) -> AsyncSQLExperimentSession:
        return AsyncSQLExperimentSession(
            self._async_session_maker.begin(),
            self._serializer,
        )

    # The following methods are required to make ExperimentSessionMaker pickleable since
    # sqlalchemy engine is not pickleable.
    # Only the engine url is pickled so the engine created upon unpickling might not be
    # exactly the same as the original one.
    def __getstate__(self):
        return {
            "url": self._engine.url,
            "async_url": self._async_engine.url,
            "serializer": self._serializer,
        }

    def __setstate__(self, state):
        engine = sqlalchemy.create_engine(state.pop("url"))
        async_engine = create_async_engine(state.pop("async_url"))
        self.__init__(state["serializer"], engine, async_engine)

    def upgrade(self) -> None:
        """Updates the database schema to the latest version.

        When called on a database already setup, this method will upgrade the database
        tables to the latest version.
        When called on an empty database, this method will create the necessary tables.
        """

        # TODO: Handle upgrading the database schema if the tables already exist.
        create_tables(self._engine)


class SQLiteExperimentSessionMaker(SQLExperimentSessionMaker):
    def __init__(
        self,
        serializer: SerializerProtocol,
        path: str,
    ):
        engine = create_engine(f"sqlite:///{path}?check_same_thread=False")
        async_engine = create_async_engine(
            f"sqlite+aiosqlite:///{path}?check_same_thread=False"
        )
        super().__init__(serializer, engine, async_engine)

    def async_session(self) -> ThreadedAsyncSQLExperimentSession:
        return ThreadedAsyncSQLExperimentSession(
            self._session_maker.begin(),
            self._serializer,
        )

    def __getstate__(self):
        return {
            "path": self._engine.url.database,
            "serializer": self._serializer,
        }

    def __setstate__(self, state):
        path = state.pop("path")
        serializer = state.pop("serializer")
        self.__init__(path, serializer)


@attrs.define
class PostgreSQLConfig:
    """Configuration for connecting to a PostgreSQL database."""

    username: str
    password: str
    host: str
    port: int
    database: str

    @classmethod
    def from_file(cls, path) -> Self:
        """Create a PostgreSQLConfig from a yaml file.

        The file should contain the same keys as the attributes of this class.
        """

        with open(path) as f:
            config = yaml.safe_load(f)
        return cls(**config)


class PostgreSQLExperimentSessionMaker(SQLExperimentSessionMaker):
    """Used to access experiment data stored in a PostgreSQL database."""

    def __init__(
        self,
        serializer: SerializerProtocol,
        config: PostgreSQLConfig,
    ):
        sync_url = URL.create(
            "postgresql+psycopg",
            username=config.username,
            password=config.password,
            host=config.host,
            port=config.port,
            database=config.database,
        )
        engine = create_engine(sync_url, isolation_level="REPEATABLE READ")
        async_url = URL.create(
            "postgresql+psycopg",
            username=config.username,
            password=config.password,
            host=config.host,
            port=config.port,
            database=config.database,
        )
        async_engine = create_async_engine(async_url, isolation_level="REPEATABLE READ")

        super().__init__(serializer, engine, async_engine)

    def async_session(self) -> ThreadedAsyncSQLExperimentSession:
        return ThreadedAsyncSQLExperimentSession(
            self._session_maker.begin(),
            self._serializer,
        )

    def __getstate__(self):
        return {
            "config": PostgreSQLConfig(
                username=self._engine.url.username,
                password=self._engine.url.password,
                host=self._engine.url.host,
                port=self._engine.url.port,
                database=self._engine.url.database,
            ),
            "serializer": self._serializer,
        }

    def __setstate__(self, state):
        config = state.pop("config")
        serializer = state.pop("serializer")
        self.__init__(serializer, config)



from collections.abc import Sequence
from typing import Optional, Literal, assert_never

import polars
from typing_extensions import deprecated

from caqtus.types.units import Unit
from ..units import with_units_added_to_columns

VALUE_FIELD = "value"
ERROR_FIELD = "error"


def is_error_dtype(dtype: polars.DataType) -> bool:
    """Check if a dtype is a QuantityDType.

    Args:
        dtype: the dtype to check.

    Returns:
        True if the dtype is a polars.Struct with two fields, magnitude and units,
        False otherwise.
    """

    if isinstance(dtype, polars.Struct):
        if len(dtype.fields) == 2:
            if (
                dtype.fields[0].name == VALUE_FIELD
                and dtype.fields[1].name == ERROR_FIELD
            ):
                return True
    return False


@polars.api.register_expr_namespace("uncertainty")
class UncertaintyExpressions:
    def __init__(self, expr: polars.Expr):
        self._expr = expr

    def nominal(self) -> polars.Expr:
        return self._expr.struct.field(VALUE_FIELD)

    def error(self) -> polars.Expr:
        return self._expr.struct.field(ERROR_FIELD)


@polars.api.register_series_namespace("uncertainty")
class UncertaintySeries:
    def __init__(self, s: polars.Series):
        self._s = s

    def nominal(self) -> polars.Series:
        """Extract the nominal value from a series containing a value and an error."""

        if not is_error_dtype(self._s.dtype):
            raise ValueError("The series must have an error dtype")
        return self._s.struct.field(VALUE_FIELD)

    def error(self) -> polars.Series:
        """Extract the error from a series containing a value and an error."""

        if not is_error_dtype(self._s.dtype):
            raise ValueError("The series must have an error dtype")
        return self._s.struct.field(ERROR_FIELD)


@deprecated("Use series.uncertainty.nominal instead")
def get_nominal_value(series: polars.Series) -> polars.Series:
    return series.uncertainty.nominal()


@deprecated("Use series.uncertainty.error instead")
def get_error(series: polars.Series) -> polars.Series:
    return series.uncertainty.error()


@polars.api.register_dataframe_namespace("uncertainty")
class UncertaintyDataframe:
    def __init__(self, df: polars.DataFrame):
        self._df = df

    def average(
        self,
        columns_to_average: Sequence[str],
        grouped_by: Sequence[str],
        error_type: Literal["sem", "std"] = "sem",
    ) -> polars.DataFrame:
        """Compute the nominal value and error of a column, grouped by hues.

        Here the nominal value is the mean of the column, and the error is either the
        standard error of the mean or the
        standard deviation.

        Args:
            columns_to_average: the name of the columns to average.
            grouped_by: the names of the columns to use to group the data. Must have at
            least one element at the moment.
            error_type: the type of error to compute. Can be "sem" (standard error of the
            mean) or "std" (standard
                deviation). Defaults to "sem".

        Returns:
            A dataframe with the columns specified in hues, plus one column with the same
            name as the column to average
            having an error dtype. The nominal value of this column is the mean of the
            column to average, and the error is
            either the standard error of the mean or the standard deviation depending on
            the value of error_type.
        """

        dataframe = self._df
        if len(grouped_by) == 0:
            raise NotImplementedError("No hue specified")

            # We convert all the y values to a single unit, so that we can compute the mean
            # and sem in a meaningful way.
            # Should add a way to select the unit to use for the averaging.
        y_magnitudes: dict[str, polars.Series] = {}
        y_units: dict[str, Optional[Unit]] = {}
        for column_to_average in columns_to_average:
            y_magnitudes[column_to_average], y_units[column_to_average] = dataframe[
                column_to_average
            ].quantity.extract_unit()

        # We need to convert all the grouped_by to a single unit, even if no operation is
        # performed on them. The issue is
        # that two values can have different magnitude and unit, but still be equal. For
        # example, 1 m and 100 cm are equal.
        # Converting to a single unit allows to avoid this problem.
        hues_magnitudes: dict[str, polars.Series] = {}
        hues_units: dict[str, Optional[Unit]] = {}
        for hue in grouped_by:
            hues_magnitudes[hue], hues_units[hue] = dataframe[
                hue
            ].quantity.extract_unit()

        dataframe_without_units = polars.DataFrame(
            [*hues_magnitudes.values(), *y_magnitudes.values()]
        )

        value_expressions: dict[str, polars.Expr] = {}
        error_expressions: dict[str, polars.Expr] = {}
        for column_to_average in columns_to_average:
            value_expressions[column_to_average] = polars.col(column_to_average).mean()
            if error_type == "sem":
                error_expressions[column_to_average] = polars.col(
                    column_to_average
                ).std() / polars.Expr.sqrt(polars.col(column_to_average).count())
            elif error_type == "std":
                error_expressions[column_to_average] = polars.col(
                    column_to_average
                ).std()
            else:
                assert_never(error_type)

        dataframe_stats_without_units = (
            dataframe_without_units.lazy()
            .group_by(*grouped_by)
            .agg(
                **{
                    column_to_average: polars.struct(
                        value_expressions[column_to_average].alias(VALUE_FIELD),
                        error_expressions[column_to_average].alias(ERROR_FIELD),
                    )
                    for column_to_average in columns_to_average
                }
            )
            .sort(*grouped_by)
            .collect()
        )

        return with_units_added_to_columns(
            dataframe_stats_without_units, {**hues_units, **y_units}
        )


def compute_stats_average(
    dataframe: polars.DataFrame,
    columns_to_average: Sequence[str],
    grouped_by: Sequence[str],
    error_type: Literal["sem", "std"] = "sem",
) -> polars.DataFrame:
    return dataframe.uncertainty.average(columns_to_average, grouped_by, error_type)



from typing import TypeGuard

from ._constant import Constant
from ._device_trigger import DeviceTrigger
from ._lane_values import LaneValues
from .compile_analog_lane import compile_analog_lane

# A channel output object is said to be a source if it generates values y(t) = f(t)
# an has no input value x(t).
ValueSource = LaneValues | DeviceTrigger | Constant


def is_value_source(obj) -> TypeGuard[ValueSource]:
    return isinstance(obj, (LaneValues, DeviceTrigger, Constant))


__all__ = [
    "LaneValues",
    "DeviceTrigger",
    "Constant",
    "ValueSource",
    "is_value_source",
    "compile_analog_lane",
]



from .sequencer import (
    Sequencer,
    ProgrammedSequence,
    SequenceStatus,
)
from ..trigger import (
    Trigger,
    SoftwareTrigger,
    ExternalTriggerStart,
    ExternalClock,
    ExternalClockOnChange,
    TriggerEdge,
)

__all__ = [
    "Sequencer",
    "Trigger",
    "SoftwareTrigger",
    "ExternalTriggerStart",
    "ExternalClock",
    "ExternalClockOnChange",
    "TriggerEdge",
    "ProgrammedSequence",
    "SequenceStatus",
]



from __future__ import annotations

import abc
import datetime
from collections.abc import Mapping, Set, Iterable
from typing import Protocol, Optional

import attrs
from returns.result import Result

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.types.data import DataLabel, Data
from caqtus.types.iteration import IterationConfiguration, Unknown
from caqtus.types.parameter import Parameter, ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from ._path import PureSequencePath
from ._path_hierarchy import PathError, PathNotFoundError
from ._state import State


class PathIsSequenceError(PathError):
    """Raised when a path is expected to be a sequence, but it is not."""

    pass


class PathIsNotSequenceError(PathError):
    """Raised when a path is expected to not be a sequence, but it is."""

    pass


class DataNotFoundError(RuntimeError):
    """Raised when data is not found in a shot."""

    pass


class SequenceStateError(RuntimeError):
    """Raised when an invalid sequence state is encountered.

    This error is raised when trying to perform an operation that is not allowed in the
    current state, such as adding data to a sequence that is not in the RUNNING state.
    """

    pass


class InvalidStateTransitionError(SequenceStateError):
    """Raised when an invalid state transition is attempted.

    This error is raised when trying to transition a sequence to an invalid state.
    """

    pass


class SequenceNotEditableError(SequenceStateError):
    """Raised when trying to edit a sequence that is not in the draft state."""

    pass


class ShotNotFoundError(RuntimeError):
    """Raised when a shot is not found in a sequence."""

    pass


class SequenceCollection(Protocol):
    """A collection of sequences inside a session.

    This abstract class defines the interface to read and write sequences in a session.
    Objects of this class provide methods for full access to read/write operations on
    sequences and their shots.
    However, they are not meant to be convenient to use directly in user code.
    Instead, consider using the higher-level API provided by the
    :class:`caqtus.session.Sequence` and :class:`caqtus.session.Shot` classes to access
    data from sequences and shots.
    """

    @abc.abstractmethod
    def is_sequence(self, path: PureSequencePath) -> Result[bool, PathNotFoundError]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_contained_sequences(self, path: PureSequencePath) -> list[PureSequencePath]:
        """Return the children of this path that are sequences, including this path.

        Return:
            A list of all sequences inside this path and all its descendants.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_global_parameters(
        self, path: PureSequencePath, parameters: ParameterNamespace
    ) -> None:
        """Set the global parameters that should be used by this sequence.

        Raises:
            SequenceNotEditable: If the sequence is not in the PREPARING state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_global_parameters(self, path: PureSequencePath) -> ParameterNamespace:
        """Get the global parameters that were used by this sequence.

        Raises:
            RuntimeError: If the sequence is in draft mode, since the global parameters
            are only set once the sequence has entered the PREPARING state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_iteration_configuration(
        self, sequence: PureSequencePath
    ) -> IterationConfiguration:
        """Return a copy of the iteration configuration for this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def set_iteration_configuration(
        self,
        sequence: PureSequencePath,
        iteration_configuration: IterationConfiguration,
    ) -> None:
        """Set the iteration configuration for this sequence.

        Raises:
            PathNotFoundError: If the path doesn't exist.
            PathIsNotSequenceError: If the path is not a sequence.
            SequenceNotEditableError: If the sequence is not in DRAFT state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_time_lanes(self, sequence_path: PureSequencePath) -> TimeLanes:
        """Return a copy of the time lanes for this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def set_time_lanes(
        self, sequence_path: PureSequencePath, time_lanes: TimeLanes
    ) -> None:
        """Set the time lanes that define how a shot is run for this sequence.

        Raises:
            PathNotFoundError: If the path doesn't exist.
            PathIsNotSequenceError: If the path is not a sequence.
            SequenceNotEditableError: If the sequence is not in DRAFT state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_device_configurations(
        self,
        path: PureSequencePath,
        device_configurations: Mapping[DeviceName, DeviceConfiguration],
    ) -> None:
        """Set the device configurations that should be used by this sequence.

        Raises:
            SequenceNotEditableError: If the sequence is not in the PREPARING state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_device_configurations(
        self, path: PureSequencePath
    ) -> Mapping[DeviceName, DeviceConfiguration]:
        """Get the device configurations that are used by this sequence.

        Raises:
            RuntimeError: If the sequence is in draft mode, since the device
            configurations are only set once the sequence has entered the PREPARING
            state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def create(
        self,
        path: PureSequencePath,
        iteration_configuration: IterationConfiguration,
        time_lanes: TimeLanes,
    ) -> None:
        """Create a new sequence at the given path.

        Raises:
            PathIsSequenceError: If the path already exists and is a sequence.
            PathHasChildrenError: If the path already exists and has children.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_state(
        self, path: PureSequencePath
    ) -> Result[State, PathNotFoundError | PathIsNotSequenceError]:
        raise NotImplementedError

    @abc.abstractmethod
    def set_state(self, path: PureSequencePath, state: State) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def get_stats(
        self, path: PureSequencePath
    ) -> Result[SequenceStats, PathNotFoundError | PathIsNotSequenceError]:
        raise NotImplementedError

    @abc.abstractmethod
    def create_shot(
        self,
        path: PureSequencePath,
        shot_index: int,
        shot_parameters: Mapping[DottedVariableName, Parameter],
        shot_data: Mapping[DataLabel, Data],
        shot_start_time: datetime.datetime,
        shot_end_time: datetime.datetime,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def get_shots(
        self, path: PureSequencePath
    ) -> Result[list[PureShot], PathNotFoundError | PathIsNotSequenceError]:
        """Return the shots that belong to this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def get_shot_parameters(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DottedVariableName, Parameter]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_all_shot_data(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DataLabel, Data]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_shot_data_by_label(
        self, path: PureSequencePath, shot_index: int, data_label: DataLabel
    ) -> Data:
        """Return the data with the given label for the shot at the given index.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
            PathIsNotSequenceError: If the path is not a sequence.
            ShotNotFoundError: If the shot does not exist in the sequence.
            DataNotFoundError: If the data with the given label does not exist in the
            shot.
        """

        raise NotImplementedError

    def get_shot_data_by_labels(
        self, path: PureSequencePath, shot_index: int, data_labels: Set[DataLabel]
    ) -> Mapping[DataLabel, Data]:
        """Return the data with the given labels for the shot at the given index.

        This method is equivalent to calling :meth:`get_shot_data_by_label` for each
        label in the set, but can be faster.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
            PathIsNotSequenceError: If the path is not a sequence.
            ShotNotFoundError: If the shot does not exist in the sequence.
            DataNotFoundError: If one of the data labels does not exist in the shot.
        """

        # Naive implementation that calls get_shot_data_by_label for each label.
        return {
            label: self.get_shot_data_by_label(path, shot_index, label)
            for label in data_labels
        }

    @abc.abstractmethod
    def get_shot_start_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        raise NotImplementedError

    @abc.abstractmethod
    def get_shot_end_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        raise NotImplementedError

    @abc.abstractmethod
    def update_start_and_end_time(
        self,
        path: PureSequencePath,
        start_time: Optional[datetime.datetime],
        end_time: Optional[datetime.datetime],
    ) -> None:
        """Update the start and end time of the sequence.

        This method is used for maintenance purposes, such as when copying a sequence
        from one session to another.
        It should not be used to record the start and end time of a sequence during
        normal operation.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_sequences_in_state(self, state: State) -> Iterable[PureSequencePath]:
        """Return all sequences in the given state."""

        raise NotImplementedError


@attrs.frozen
class SequenceStats:
    state: State
    start_time: Optional[datetime.datetime]
    stop_time: Optional[datetime.datetime]
    number_completed_shots: int
    expected_number_shots: int | Unknown


@attrs.frozen
class PureShot:
    """Unique identifier for a shot in a sequence."""

    sequence_path: PureSequencePath
    index: int



"""Defines image and region of interest types."""

from . import roi
from ._image_type import Image, is_image, ImageLabel, Width, Height

__all__ = [
    "Image",
    "is_image",
    "ImageLabel",
    "Width",
    "Height",
    "roi",
]



from __future__ import annotations

from typing import Optional

import anyio.to_thread
import attrs
import numpy as np
import polars
import pyqtgraph
import qtawesome
from PySide6.QtCore import QStringListModel
from PySide6.QtWidgets import QWidget, QVBoxLayout, QCompleter
from caqtus.analysis.units import extract_unit
from caqtus.gui.graphplot.views.view import DataView

from .scatter_view_ui import Ui_ScatterView


class ScatterView(DataView, Ui_ScatterView):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setupUi(self)

        self.plot_widget = pyqtgraph.PlotWidget(self, background="white")
        self.plot_widget.enableAutoRange()
        self.plot_item = self.plot_widget.getPlotItem()
        self.scatter_plot = pyqtgraph.ScatterPlotItem()
        self.plot_item.addItem(self.scatter_plot)

        layout = self.layout()
        assert isinstance(layout, QVBoxLayout)
        layout.insertWidget(1, self.plot_widget)
        self.settings_button.setIcon(qtawesome.icon("mdi6.cog"))
        self.apply_button.setIcon(qtawesome.icon("mdi6.check"))
        self.apply_button.clicked.connect(self.on_apply)

        self.clear()

        self.x_column: Optional[str] = None
        self.y_column: Optional[str] = None

        self.columns_model = QStringListModel(self)
        self.columns_completer = QCompleter(self.columns_model, self)
        self.x_line_edit.setCompleter(self.columns_completer)
        self.y_line_edit.setCompleter(self.columns_completer)

    def on_apply(self) -> None:
        x_column = self.x_line_edit.text()
        y_column = self.y_line_edit.text()
        self.x_column = x_column
        self.y_column = y_column

    def clear(self) -> None:
        self.scatter_plot.setData([], [])

    async def update_data(self, data: polars.DataFrame) -> None:
        column_names = data.columns
        if column_names != self.columns_model.stringList():
            # We only reset the completer if the columns have actually changed,
            # otherwise we would reset the user input.
            self.columns_model.setStringList(data.columns)
        if self.x_column is None or self.y_column is None:
            self.clear()
            return
        if data.is_empty():
            self.clear()
            return
        to_plot = await anyio.to_thread.run_sync(
            self.update_plot, self.x_column, self.y_column, data
        )
        self.scatter_plot.setData(to_plot.x_values, to_plot.y_values)
        self.plot_item.setLabel("bottom", to_plot.x_label)
        self.plot_item.setLabel("left", to_plot.y_label)

    @staticmethod
    def update_plot(x_column: str, y_column: str, data: polars.DataFrame) -> PlotInfo:
        x_series = data[x_column]
        x_magnitude, x_unit = extract_unit(x_series)
        y_series = data[y_column]
        y_magnitude, y_unit = extract_unit(y_series)

        if x_unit:
            x_label = f"{x_column} [{x_unit:~}]"
        else:
            x_label = x_column
        if y_unit:
            y_label = f"{y_column} [{y_unit:~}]"
        else:
            y_label = y_column

        plot_info = PlotInfo(
            x_values=np.array(x_magnitude),
            y_values=np.array(y_magnitude),
            x_range=(float(x_magnitude.min()), float(x_magnitude.max())),
            y_range=(float(y_magnitude.min()), float(y_magnitude.max())),
            x_label=x_label,
            y_label=y_label,
        )
        return plot_info


@attrs.define
class PlotInfo:
    x_values: np.ndarray
    y_values: np.ndarray
    x_range: tuple[float, float]
    y_range: tuple[float, float]
    x_label: str
    y_label: str



from __future__ import annotations

from collections.abc import Iterable
from typing import Optional

import attrs
from PySide6.QtCore import (
    QModelIndex,
    Qt,
    QObject,
    QMimeData,
)
from PySide6.QtGui import QStandardItem, QStandardItemModel, QPalette

from caqtus.types.expression import Expression
from caqtus.types.iteration import (
    StepsConfiguration,
    Step,
    ExecuteShot,
    VariableDeclaration,
    LinspaceLoop,
    ArangeLoop,
)
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils import serialization

NAME_COLOR = "#AA4926"
VALUE_COLOR = "#6897BB"
HIGHLIGHT_COLOR = "#cc7832"


@attrs.define
class ExecuteShotData:
    def display_data(self) -> str:
        return f"<span style='color:{HIGHLIGHT_COLOR}'>do shot</span>"


@attrs.define
class VariableDeclarationData:
    variable: DottedVariableName
    value: Expression

    def display_data(self) -> str:
        text_color = f"#{QPalette().text().color().rgba():X}"
        return (
            f"<span style='color:{NAME_COLOR}'>{self.variable}</span> "
            f"<span style='color:{text_color}'>=</span> "
            f"<span style='color:{VALUE_COLOR}'>{self.value}</span>"
        )


@attrs.define
class LinspaceLoopData:
    variable: DottedVariableName
    start: Expression
    stop: Expression
    num: int

    def display_data(self) -> str:
        text_color = f"#{QPalette().text().color().rgba():X}"
        return (
            f"<span style='color:{HIGHLIGHT_COLOR}'>for</span> "
            f"<span style='color:{NAME_COLOR}'>{self.variable}</span> "
            f"<span style='color:{text_color}'>=</span> "
            f"<span style='color:{VALUE_COLOR}'>{self.start}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>to </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.stop}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>with </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.num}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>steps:</span>"
        )


@attrs.define
class ArrangeLoopData:
    variable: DottedVariableName
    start: Expression
    stop: Expression
    step: Expression

    def display_data(self) -> str:
        text_color = f"#{QPalette().text().color().rgba():X}"
        return (
            f"<span style='color:{HIGHLIGHT_COLOR}'>for</span> "
            f"<span style='color:{NAME_COLOR}'>{self.variable}</span> "
            f"<span style='color:{text_color}'>=</span> "
            f"<span style='color:{VALUE_COLOR}'>{self.start}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>to </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.stop}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>with </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.step}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>spacing:</span>"
        )


StepData = (
    ExecuteShotData | VariableDeclarationData | LinspaceLoopData | ArrangeLoopData
)


class StepItem(QStandardItem):
    @classmethod
    def construct(cls, step: Step) -> StepItem:
        item = cls()
        flags = (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsDragEnabled
        )
        match step:
            case ExecuteShot():
                item.setData(ExecuteShotData(), Qt.ItemDataRole.EditRole)
                item.setFlags(flags | Qt.ItemFlag.ItemNeverHasChildren)
            case VariableDeclaration(variable=variable, value=value):
                item.setData(
                    VariableDeclarationData(variable=variable, value=value),
                    Qt.ItemDataRole.EditRole,
                )
                item.setFlags(
                    flags
                    | Qt.ItemFlag.ItemIsEditable
                    | Qt.ItemFlag.ItemNeverHasChildren
                )
            case LinspaceLoop(
                variable=variable, start=start, stop=stop, num=num, sub_steps=sub_steps
            ):
                children = [cls.construct(sub_step) for sub_step in sub_steps]
                item.setData(
                    LinspaceLoopData(
                        variable=variable, start=start, stop=stop, num=num
                    ),
                    Qt.ItemDataRole.EditRole,
                )
                for child in children:
                    item.appendRow(child)
                item.setFlags(
                    flags | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDropEnabled
                )
            case ArangeLoop(
                variable=variable,
                start=start,
                stop=stop,
                step=step,
                sub_steps=sub_steps,
            ):
                children = [cls.construct(sub_step) for sub_step in sub_steps]
                item.setData(
                    ArrangeLoopData(
                        variable=variable, start=start, stop=stop, step=step
                    ),
                    Qt.ItemDataRole.EditRole,
                )
                for child in children:
                    item.appendRow(child)
                item.setFlags(
                    flags | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDropEnabled
                )
            case _:
                raise NotImplementedError(f"Step {step} not supported")

        return item

    def data(self, role: Qt.ItemDataRole.DisplayRole = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            step_data = self.data(Qt.ItemDataRole.EditRole)
            assert isinstance(step_data, StepData)
            return step_data.display_data()
        return super().data(role)

    def get_step(self) -> Step:
        data = self.data(role=Qt.ItemDataRole.EditRole)
        match data:
            case ExecuteShotData():
                return ExecuteShot()
            case VariableDeclarationData(variable=variable, value=value):
                return VariableDeclaration(variable=variable, value=value)
            case LinspaceLoopData(variable=variable, start=start, stop=stop, num=num):
                child_items = [self.child(i) for i in range(self.rowCount())]
                sub_steps = [item.get_step() for item in child_items]
                return LinspaceLoop(
                    variable=variable,
                    start=start,
                    stop=stop,
                    num=num,
                    sub_steps=sub_steps,
                )
            case ArrangeLoopData(variable=variable, start=start, stop=stop, step=step):
                child_items = [self.child(i) for i in range(self.rowCount())]
                sub_steps = [item.get_step() for item in child_items]
                return ArangeLoop(
                    variable=variable,
                    start=start,
                    stop=stop,
                    step=step,
                    sub_steps=sub_steps,
                )
            case _:
                raise NotImplementedError(f"Step {data} not supported")


class StepsModel(QStandardItemModel):
    def __init__(self, steps: StepsConfiguration, parent: Optional[QObject] = None):
        super().__init__(parent)
        self.set_steps(steps)
        self._read_only = True

    def set_read_only(self, read_only: bool):
        self._read_only = read_only

    def is_read_only(self) -> bool:
        return self._read_only

    def get_steps(self) -> StepsConfiguration:
        root = self.invisibleRootItem()
        items = [root.child(i) for i in range(root.rowCount())]
        assert all(isinstance(item, StepItem) for item in items)
        steps = [item.get_step() for item in items]
        return StepsConfiguration(steps=steps)

    def set_steps(self, steps: StepsConfiguration):
        self.beginResetModel()
        self.clear()
        items = [StepItem.construct(step) for step in steps.steps]
        root = self.invisibleRootItem()
        for item in items:
            root.appendRow(item)
        self.endResetModel()

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if self._read_only:
            return False
        return super().setData(index, value, role)

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        flags = super().flags(index)

        if self._read_only:
            flags &= ~Qt.ItemFlag.ItemIsEditable
            flags &= ~Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def supportedDropActions(self) -> Qt.DropAction:
        return Qt.DropAction.MoveAction

    def mimeTypes(self) -> list[str]:
        return ["application/json"]

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData:
        items = [self.itemFromIndex(index) for index in indexes]
        assert all(isinstance(item, StepItem) for item in items)
        descendants = []
        for item in items:
            descendants.extend(get_strict_descendants(item))
        items = [item for item in items if item not in descendants]

        steps = [item.get_step() for item in items]
        serialized = serialization.to_json(steps)
        mime_data = QMimeData()
        mime_data.setData("application/json", serialized.encode())
        mime_data.setText(serialized)
        return mime_data

    def canDropMimeData(self, data, action, row: int, column: int, parent: QModelIndex):
        if self._read_only:
            return False
        if not data.hasFormat("application/json"):
            return False
        return super().canDropMimeData(data, action, row, column, parent)

    def dropMimeData(
        self,
        data: QMimeData,
        action: Qt.DropAction,
        row: int,
        column: int,
        parent: QModelIndex,
    ) -> bool:
        if not self.canDropMimeData(data, action, row, column, parent):
            return False

        json_string = data.data("application/json").data().decode()
        try:
            steps = serialization.from_json(json_string, list[Step])
        except ValueError:
            return False

        new_items = [StepItem.construct(step) for step in steps]

        parent_item = (
            self.itemFromIndex(parent) if parent.isValid() else self.invisibleRootItem()
        )
        if not (parent_item.flags() & Qt.ItemFlag.ItemIsDropEnabled):
            return False
        if row == -1:
            row = parent_item.rowCount()
        parent_item.insertRows(row, new_items)
        return True

    def insert_above(self, step: Step, index: QModelIndex):
        if self._read_only:
            return
        if not index.isValid():
            return
        else:
            parent = index.parent()
            parent_item = (
                self.itemFromIndex(parent)
                if parent.isValid()
                else self.invisibleRootItem()
            )
            new_item = StepItem.construct(step)
            parent_item.insertRows(index.row(), [new_item])


def get_strict_descendants(parent: QStandardItem) -> list[QStandardItem]:
    children = [parent.child(i) for i in range(parent.rowCount())]
    descendants = []
    descendants.extend(children)
    for child in children:
        descendants.extend(get_strict_descendants(child))
    return descendants
